from mlx_lm.models.cache import KVCache, RotatingKVCache
import mlx.nn as nn
import mlx.core as mx
from mlx.utils import tree_flatten, tree_unflatten
from typing import List, Optional, Any, Dict, Union, Tuple, NamedTuple

class CacheHistory:
    def __init__(self,
            cache: List[Union[KVCache, RotatingKVCache]],
            token_ids: List[List[int]]
        ) -> None:
        self.cache: List[Union[KVCache, RotatingKVCache]] = cache
        self.token_ids: List[List[int]] = token_ids


def make_empty_cache(model: nn.Module, max_kv_size: Optional[int] = None) -> List[Union[KVCache, RotatingKVCache]]:
    """
    Constructs the model's cache for use during generation.

    This function defers the cache construction to the model if it has a
    ``make_cache`` method. If the model does not have this method, a default
    KV cache is created. If a maximum key-value size is specified and the model
    does not have a ``make_cache`` method, a ``RotatingKVCache`` is used with
    the specified maximum size.
    Args:
        model (nn.Module): The language model.
        max_kv_size (Optional[int]): If provided and the model does not have a
            ``make_cache`` method, a ``RotatingKVCache`` is used with a maximum
            size of ``max_kv_size``.

    Returns:
        List[Union[KVCache, RotatingKVCache]]: The constructed cache.
    """
    if hasattr(model, "make_cache"):
        return model.make_cache()

    num_layers = len(model.layers)
    if max_kv_size is not None:
        return [
            RotatingKVCache(max_size=max_kv_size, keep=4) for _ in range(num_layers)
        ]
    else:
        return [KVCache() for _ in range(num_layers)]
    
def create_cache_dict(cache: List[Union[KVCache, RotatingKVCache]]) -> Dict[str, mx.array]:
    """Create a dictionary of prompt cache for saving or further manipuation. 

    Args:
        cache (List[Union[KVCache, RotatingKVCache]]): Prompt cache.

    Returns:
        Dict[str, mx.array]: A dictionary of prompt cache for saving or further manipuation. 
    """
    offset = cache[0].offset
    if offset != 0:
        cache_dict = dict(tree_flatten([c.state for c in cache]))
        return cache_dict
    
def save_cache(cache: List[Union[KVCache, RotatingKVCache]], file: str, metadata: Optional[Dict[str, str]] = None) -> None:
    """Save cache as a safetensors file.

    Args:
        cache (List[Union[KVCache, RotatingKVCache]]): Prompt cache to save.
        file (str): File the cache should be saved.
        metadata (Optional[Dict[str, str]], optional): Metadata for the prompt cache. Defaults to None.
    """
    cache_dict = create_cache_dict(cache=cache)
    if cache_dict:
        mx.save_safetensors(file=file, arrays=cache_dict, metadata=metadata)
    del cache_dict
    mx.metal.clear_cache()

def load_cache(file: str, max_kv_size: Optional[int] = None) -> Tuple[List[Union[KVCache, RotatingKVCache]], Dict[str, str]]:
    """Load a cache file.

    Args:
        file (str): Safetensors file to load.
        max_kv_size (Optional[int], optional): Max number of tokens being kept in the cache. If None given, there is no limitation. Defaults to None.

    Returns:
        Tuple[List[Union[KVCache, RotatingKVCache]], Dict[str, str]]: Cache and it's metadata.
    """
    cache_dict, metadata = mx.load(file, return_metadata=True)
    cache_list = tree_unflatten(list(cache_dict.items()))
    cache = []
    for i, (key, value) in enumerate(cache_list):
        cache.append(RotatingKVCache(max_size=max_kv_size, keep=4) if max_kv_size else KVCache())
        cache[i].update_and_fetch(key, value)
    mx.eval([c.state for c in cache])
    del cache_dict
    mx.metal.clear_cache()
    return cache, metadata

def split_cache(cache: List[Union[KVCache, RotatingKVCache]]) -> List[List[KVCache]]:
    """Split the cache into multiple smaller caches. Each smaller cache will contain a subset of the original cache's data.

    Args:
        cache (List[Union[KVCache, RotatingKVCache]]): The cache to be split.

    Returns:
        List[List[KVCache]]: A list of smaller caches.
    """
    num_split = cache[0].state[0].shape[0]
    caches = []
    def make_cache(n):
        ec = []
        for j, c in enumerate(cache):
            ec.append(KVCache())
            ec[j].update_and_fetch(c.state[0][n:n + 1], c.state[1][n:n + 1])
            mx.eval(ec[j].state)
        return ec
    for i in range(num_split):
        caches.append(make_cache(i))
    del cache
    mx.metal.clear_cache()
    return caches

def trim_cache(cache: List[KVCache], offset: int, trim_suffix: int = 0) -> List[KVCache]:
    """
    Trim the cache by removing entries before the specified offset and optionally
    removing entries after the specified trim_suffix.

    Args:
        cache (List[KVCache]): The cache to be trimmed.
        offset (int): The offset from which to start trimming.
        trim_suffix (int, optional): The number of entries after the offset to keep. Defaults to 0.

    Returns:
        List[KVCache]: The trimmed cache.
    """
    new_cache = []
    for j, c in enumerate(cache):
        new_cache.append(KVCache())
        if trim_suffix == 0:
            new_cache[j].update_and_fetch(c.state[0][:, :, offset:, :], c.state[1][:, :, offset:, :])
        else:
            new_cache[j].update_and_fetch(c.state[0][:, :, offset:-trim_suffix, :], c.state[1][:, :, offset:-trim_suffix, :])
        mx.eval(new_cache[j].state)
    del cache
    mx.metal.clear_cache()
    return new_cache


def select_from_cache(
        cache: List[Union[KVCache, RotatingKVCache]],
        prompt_index: int = 0,
        select_len: Optional[int] = None,
        start_from: int = 0
    ) -> List[Tuple[mx.array, mx.array]]:
    """Get the kv cache tensors of a specific sequence in a cache with multiple prompt cache sequences.

    Args:
        cache (List[Union[KVCache, RotatingKVCache]]): Cache object.
        prompt_index (int, optional): The index of the sequence to get. Defaults to 0.
        select_len (Optional[int], optional): Length of the seqence to get. Defaults to None.
        start_from (int, optional): Starting index of the sequence. Defaults to 0.

    Returns:
        List[Tuple[mx.array, mx.array]]: List of keys and values tuples for updating the new cache.
    """
    num_prompts, num_kv_heads, prompt_len, head_dim = cache[0].state[0].shape
    max_index = select_len + start_from if select_len else prompt_len
    if prompt_index >= num_prompts:
        raise ValueError('Prompt index out of index.')
    if max_index > prompt_len:
        raise ValueError(f'Select length out of index.')
    cache_list = []
    for c in cache:
        k, v  = c.state
        cache_list.append((k[prompt_index:prompt_index + 1, :, start_from:max_index, :], v[prompt_index:prompt_index + 1, :, start_from:max_index, :]))
    return cache_list

def find_max_prefix_num(new: List[int], baseline: List[int]) -> int:
    """Helper function to find the maximum number of tokens shared in the prefix of two prompts.

    Args:
        new (List[int]): First prompt token ids.
        baseline (List[int]): Second prompt token ids.

    Returns:
        int: The maximum number of tokens shared in the prefix of two prompts.
    """
    from itertools import takewhile
    return len(list(takewhile(lambda x: x[0] == x[1], zip(new, baseline))))



import uvicorn
from fastapi import FastAPI, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse, StreamingResponse
from .engine import ModelEngine, TextCompletionInput, ChatCompletionInput
from .model_utils import PACKAGE_NAME, ModelConfig
import asyncio
import argparse
import logging, json, warnings, yaml
from typing import Union, Dict, Any, List, Tuple, Callable, Optional

warnings.filterwarnings('ignore')
logging.basicConfig(format='[(%(levelname)s) %(asctime)s]: %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
logger = logging.getLogger(__name__)

# Configure
def get_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=f'{PACKAGE_NAME}.server',
        description='Run an OpenAI-compatible LLM server.',
    )
    parser.add_argument('-m', '--model-path', type=str, 
        default=None, help='Path to the model or the HuggingFace repository name if only one model should be served.')
    parser.add_argument('--tokenizer-path', type=str, 
        default=None, help='Path to the tokenizer or the HuggingFace repository name if only one model should be served. If None is given, it will be the model_path. Defaults to None.')
    parser.add_argument('--adapter-path', type=str, 
        default=None, help='Path to the adapter for the model. Defaults to None.')
    parser.add_argument('--revision', type=str, 
        default=None, help='Rivision of the repository if an HF repository is given. Defaults to None.')
    parser.add_argument('-q', '--quantize', type=str, 
        default='fp16', help='Model qunatization, options are "fp16", "q8", "q4", "q2". Defaults to "fp16", meaning no quantization.')
    parser.add_argument('--model-name', type=str,
        default=None, help='Model name appears in the API endpoint. If None is given, it will be created automatically with the model path. Defaults to None.')
    parser.add_argument('-cf', '--config-file', type=str, 
        default=None, 
        help='Path of the config file that store the configs of all models wanted to be served. If this is passed, "model-path", "quantize", and "model-name" will be ignored.')
    parser.add_argument('--prefill-step-size', type=int, 
        default=512, help='Batch size for model prompt processing. Defaults to 512.')
    parser.add_argument('-mk', '--max-keep', type=int, 
        default=50, help='Maximum number of cache history for each model to keep. Defaults to 50.')
    parser.add_argument('--token-threshold', type=int,
        default=20, 
        help='Minimum number of tokens in the prompt plus generated text to trigger prompt caching. Shorter prompts do not require caching to speed up generation. Defaults to 20.')
    parser.add_argument('--api-key', type=str, default=None, help='API key to access the endpoints. Defaults to None.')
    parser.add_argument('-p', '--port', type=int, 
                        default=5001, help='Port to server the API endpoints.')
    parser.add_argument('--host', type=str, 
                        default='127.0.0.1', help='Host to bind the server to. Defaults to "127.0.0.1".')
    return parser

def parse_args() -> Tuple[ModelEngine, int, List[str]]:
    engine_args = get_arg_parser().parse_args()
    host = engine_args.host
    port = engine_args.port
    api_keys = [engine_args.api_key] if engine_args.api_key is not None else []
    if engine_args.config_file is not None:
        with open(engine_args.config_file, 'r') as f:
            model_args = yaml.safe_load(f)
        if isinstance(model_args, list):
            model_args = [ModelConfig(**args) for args in model_args]
        else:
            model_args = [ModelConfig(**model_args)]
    elif engine_args.model_path is not None:
        model_args = [
            ModelConfig(
                model_id_or_path=engine_args.model_path,
                tokenizer_id_or_path=engine_args.tokenizer_path,
                adapter_path=engine_args.adapter_path,
                quant=engine_args.quantize,
                revision=engine_args.revision,
                model_name=engine_args.model_name
            )
        ]
    else:
        raise ValueError('Either model_path or config_file has to be provide.')
    engine = ModelEngine(models=model_args, prefill_step_size=engine_args.prefill_step_size, 
        token_threshold=engine_args.token_threshold, max_keep=engine_args.max_keep, logger=logger)
    return engine, host, port, api_keys

def convert_arguments(new: str, old: str, args: Dict[str, Any], transform_fn: Optional[Callable] = None) -> Dict[str, Any]:
    value = args.pop(new, None)
    if value is not None:
        args[old] = value if transform_fn is None else transform_fn(value)
    return args

engine, host, port, api_keys = parse_args()

app = FastAPI()
semaphore = asyncio.Semaphore(1)
    
    
async def async_generate_stream(args: Dict[str, Any]):
    args['completion_type'] = 'text_completion'
    generator = await asyncio.to_thread(engine.generate, **args)
    try:
        for token in generator:
            yield f'data: {json.dumps(token)}\n\n'
        yield 'data: [DONE]'
    except Exception as e:
        logger.error(e)
        raise e

async def async_generate(args: Dict[str, Any]):
    args['completion_type'] = 'text_completion'
    result = await asyncio.to_thread(engine.generate, **args)
    return result

async def async_chat_generate_stream(args: Dict[str, Any]):
    generator = await asyncio.to_thread(engine.chat_generate, **args)
    try:
        for token in generator:
            yield f'data: {json.dumps(token)}\n\n'
        yield 'data: [DONE]'
    except asyncio.CancelledError:
        logger.error('Stopped in chat stream.')

async def async_chat_generate(args: Dict[str, Any]):
    result = await asyncio.to_thread(engine.chat_generate, **args)
    return result

@app.post('/v1/completions',  response_model=None)
async def completions(request: Request) -> Union[StreamingResponse, JSONResponse]:
    content = await request.json()
    api_key = request.headers.get('authorization', 'Bearer ').removeprefix('Bearer ')
    if ((api_key not in api_keys) and (len(api_keys) != 0)):
        return JSONResponse(jsonable_encoder(dict(error='Invalid API key.')), status_code=404)
    model = content.get('model')
    if model not in engine.models.keys():
        return JSONResponse(jsonable_encoder(dict(error=f'Model "{model}" does not exist.')), status_code=404)
    stream = content.get('stream', False)
    content = convert_arguments('max_completion_tokens', 'max_tokens', content)
    content = convert_arguments('frequency_penalty', 'repetition_penalty', content)
    content = convert_arguments('response_format', 'guided_json', content, transform_fn=lambda x: x.get('json_schema', dict()).get('schema'))
    if isinstance(content.get('stop', None), str):
        content['stop'] = [content['stop']]
    args_model = TextCompletionInput(**content)
    logger.info(args_model)
    args = args_model.model_dump()
    async with semaphore:
        if stream:
            return StreamingResponse(async_generate_stream(args), media_type="text/event-stream")
        else:
            result = await async_generate(args)
            return JSONResponse(jsonable_encoder(result))
        
@app.post('/v1/chat/completions',  response_model=None)
async def completions(request: Request) -> Union[StreamingResponse, JSONResponse]:
    content = await request.json()
    api_key = request.headers.get('authorization', 'Bearer ').removeprefix('Bearer ')
    if ((api_key not in api_keys) and (len(api_keys) != 0)):
        return JSONResponse(jsonable_encoder(dict(error='Invalid API key.')), status_code=404)
    model = content.get('model')
    if model not in engine.models.keys():
        return JSONResponse(jsonable_encoder(dict(error=f'Model "{model}" does not exist.')), status_code=404)
    stream = content.get('stream', False)
    content = convert_arguments('max_completion_tokens', 'max_tokens', content)
    content = convert_arguments('frequency_penalty', 'repetition_penalty', content)
    content = convert_arguments('response_format', 'guided_json', content, transform_fn=lambda x: x.get('json_schema', dict()).get('schema'))
    if isinstance(content.get('stop', None), str):
        content['stop'] = [content['stop']]
    args_model = ChatCompletionInput(**content)
    logger.info(args_model)
    args = args_model.model_dump()
    async with semaphore:
        if stream:
            return StreamingResponse(async_chat_generate_stream(args), media_type="text/event-stream")
        else:
            result = await async_chat_generate(args)
            return JSONResponse(jsonable_encoder(result))

@app.get('/v1/models')
async def get_models(request: Request) -> JSONResponse:
    api_key = request.headers.get('authorization', 'Bearer ').removeprefix('Bearer ')
    if ((api_key not in api_keys) and (len(api_keys) != 0)):
        return JSONResponse(jsonable_encoder(dict(error='Invalid API key.')), status_code=404)
    return JSONResponse(content=jsonable_encoder(dict(object='list', data=engine.model_info))) 

@app.get('/v1/models/{model_id}')
async def get_model(request: Request, model_id: str) -> JSONResponse:
    api_key = request.headers.get('authorization', 'Bearer ').removeprefix('Bearer ')
    if ((api_key not in api_keys) and (len(api_keys) != 0)):
        return JSONResponse(jsonable_encoder(dict(error='Invalid API key.')), status_code=404)
    model_dict = {info['id']: info for info in engine.model_info}
    if model_id not in model_dict.keys():
        return JSONResponse(jsonable_encoder(dict(error='Invalid model ID.')), status_code=404)
    return JSONResponse(content=jsonable_encoder(model_dict[model_id]))

def main():
    uvicorn.run(app, port=port, host=host)

if __name__ == '__main__':
    main()



from mlx_lm.tokenizer_utils import TokenizerWrapper
from mlx.nn import Module
from datetime import datetime as dt
import uuid
import mlx.core as mx
from .cache_utils import CacheHistory, make_empty_cache
from .model_utils import package_cache_dir, get_model_and_tokenizer, ModelConfig, get_model_name, make_model_exist
from .generate_utils import (
    generate, 
    stream_generate, 
    remove_bos_duplicates,
    get_choice_processor, 
    get_json_processor, 
    get_regex_processor, 
    get_grammar_processor,
    GenerationOutput, 
    OUTLINE_INSTALLED
    )
from .chat_utils import ChatTemplate, convert_tool_to_json_schema, get_tool_name
from .caching import CacheManager
from logging import Logger
from pydantic import BaseModel, Field, field_validator
import os
from typing import Optional, List, Dict, Union, Literal, Any, Iterator, Generator
if OUTLINE_INSTALLED:
    from outlines.models.transformers import TransformerTokenizer
    from outlines.processors.base_logits_processor import OutlinesLogitsProcessor

def get_return_dict(
        generation_outputs: List[GenerationOutput], 
        id: str, model: str, 
        completion_type: Literal['text_completion', 'chat.completion'], 
        created: float,
        stream: bool,
        indices: Optional[List[int]] = None,
        tool_start: Optional[str] = None,
        tool_call_id: Optional[str] = None,
        tool_name: Optional[str] = None,
        tool_arguments: Optional[str] = None
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    num_outputs = len(generation_outputs)
    indices = indices if indices else range(num_outputs)
    if num_outputs != len(indices):
        raise Exception('Indices length mistmatch.')
    if completion_type == 'text_completion':
        choices = [dict(index=i, finish_reason=go.finish_reason if go.stop_text != tool_start else 'tool_call_start', text=go.text) for i, go in zip(indices, generation_outputs)]
        if stream:
            out_dicts = [dict(
            id='cmpl-'+id, 
            object=completion_type, 
            created=created,
            model=model,
            choices=[c]
            ) for c in choices]
        else:
            tokens = [len(go.token_ids.tolist()) for go in generation_outputs]
            out_dicts = dict(id='cmpl-'+id, 
            object=completion_type, 
            created=created,
            model=model, 
            choices=choices,
            usage=dict(
                total_tokens=sum(tokens)
            )
        )
        return out_dicts
    elif ((completion_type == 'chat.completion') and (stream)):
        return [dict(
            id='chatcmpl-'+id, 
            object=completion_type + '.chunk', 
            created=created, 
            model=model, 
            choices=[
                dict(
                    index=i, 
                    finish_reason=(go.finish_reason if go.stop_text != tool_start else 'tool_call_start') if tool_call_id is None else 'tool_calls', 
                    delta=dict(
                        role="assistant",
                        content=go.text if tool_call_id is None else None,
                        tool_calls=[] if tool_call_id is None else [
                                {
                                    "index": 0,
                                    "id": f'call_{tool_call_id}',
                                    "function": {
                                        "arguments": tool_arguments,
                                        "name": tool_name
                                    },
                                    "type": 'function'
                                }
                            ]
                    )
                )
            ]
        ) for i, go in zip(indices, generation_outputs)]
    elif completion_type == 'chat.completion':
        tokens = [len(go.token_ids.tolist()) for go in generation_outputs]
        return dict(
            id='chatcmpl-'+id, 
            object=completion_type, 
            created=created, 
            model=model, 
            usage=dict(
                total_tokens=sum(tokens)
            ), 
            choices=[
                dict(
                    index=i, 
                    message=dict(
                        role="assistant", 
                        content=go.text if tool_call_id is None else None,
                        tool_calls=[] if tool_call_id is None else [
                                {
                                    "index": 0,
                                    "id": f'call_{tool_call_id}',
                                    "function": {
                                        "arguments": tool_arguments,
                                        "name": tool_name
                                    },
                                    "type": 'function'
                                }
                            ]
                    ),
                    finish_reason=(go.finish_reason if go.stop_text != tool_start else 'tool_call_start') if tool_call_id is None else 'tool_calls'
                )
            for i, go in enumerate(generation_outputs)]
        )

class TextCompletionInput(BaseModel):
    model: str
    prompt: Union[str, List[str]]
    stream: bool = False
    stop: Optional[List[str]] = None
    max_tokens: int = Field(default=4096, ge=1, le=32000)
    n: int = Field(default=1, ge=1)
    temperature: float = Field(default=0.0, ge=0.0, le=2.0)
    repetition_penalty: Optional[float] = Field(default=None, gt=0.0, le=2.0)
    repetition_context_size: Optional[int] = Field(default=100, ge=1)
    top_p: float = Field(default=1.0, gt=0.0, le=1.0)
    min_p: float = Field(default=0.0, ge=0.0, le=1.0)
    min_tokens_to_keep: int = Field(default=1, ge=1)
    logit_bias: Optional[Dict[int, float]] = None
    seed: Optional[int] = None
    guided_json: Optional[Union[str, dict]] = None
    guided_choice: Optional[List[str]] = None
    guided_regex: Optional[str] = None
    guided_grammar: Optional[str] = None
    guided_whitespace_pattern: Optional[str] = None

    @field_validator('prompt')
    def check_prompt(cls, v):
        if isinstance(v, str) and not v.strip():
            raise ValueError('Prompt cannot be an empty string.')
        if isinstance(v, list) and not v:
            raise ValueError('Prompt cannot be an empty list.')
        return v

class ChatCompletionInput(BaseModel):
    model: str
    messages: Union[List[Dict[str, Any]], List[List[Dict[str, Any]]]]
    stream: bool = False
    stop: Optional[List[str]] = None
    max_tokens: int = Field(default=4096, ge=1, le=32000)
    n: int = Field(default=1, ge=1)
    temperature: float = Field(default=0.0, ge=0.0, le=2.0)
    repetition_penalty: Optional[float] = Field(default=None, gt=0.0, le=2.0)
    repetition_context_size: Optional[int] = Field(default=100, ge=1)
    top_p: float = Field(default=1.0, gt=0.0, le=1.0)
    min_p: float = Field(default=0.0, ge=0.0, le=1.0)
    min_tokens_to_keep: int = Field(default=1, ge=1)
    logit_bias: Optional[Dict[int, float]] = None
    seed: Optional[int] = None
    guided_json: Optional[Union[str, dict]] = None
    guided_choice: Optional[List[str]] = None
    guided_regex: Optional[str] = None
    guided_grammar: Optional[str] = None
    guided_whitespace_pattern: Optional[str] = None
    tools: Optional[List[Dict[str, Any]]] = None
    tool_choice: Union[Literal['none', 'auto', 'required'], Dict[str, Union[str, Dict[str, str]]]] = 'auto'

    @field_validator('messages')
    def check_messages(cls, v):
        if isinstance(v, list) and len(v) == 0:
            raise ValueError('Messages cannot be an empty list.')
        for msgs in v:
            if len(msgs) == 0:
                raise ValueError('Messages cannot be an empty list.')
        return v

class ModelEngine:

    def __init__(self, 
            models: Union[ModelConfig, List[ModelConfig]],
            prefill_step_size: int = 512,
            token_threshold: int = 20,
            save_similarity_threshold: float = 0.9,
            max_keep: int = 50, 
            logger: Optional[Logger] = None
        ) -> None:
        """Initialising the engine.

        Args:
            models (Union[ModelConfig, List[ModelConfig]]): Model configurations or list of model configurations to serve.
            prefill_step_size (int, optional): Batch size for prompt preprocessing. Defaults to 512.
            token_threshold (int, optional): Minimum number of tokens to be considered to save as a prompt cache history. Defaults to 20.
            max_keep (int, optional): Maximum number of cache history files to keep. Defaults to 50.
            logger (Optional[Logger], optional): If a logger is provided, messages of adding and deleting caches will be added.
        """
        self.model_configs = [models] if isinstance(models, ModelConfig) else models
        self.prefill_step_size = prefill_step_size
        self.token_threshold = token_threshold
        self.save_sim_threshold = save_similarity_threshold
        self.max_keep = max_keep
        self.logger = logger
        self.models: Dict[str, ModelConfig] = dict()
        self.model: Optional[Module] = None
        self.tokenizer: Optional[TokenizerWrapper] = None
        self.current_model: Optional[str] = None
        self.cache_manager: Optional[CacheManager] = None
        self.outlines_tokenizer: Optional[TransformerTokenizer] = None
        self.chat_template: Optional[ChatTemplate] = None
        self.cache_dir = os.path.join(package_cache_dir, 'prompt_cache')
        for model_config in self.model_configs:
            self._prepare_model(model_config=model_config)
        self.model_info = [dict(id=m, object='model', created=int(dt.now().timestamp()), owned_by=None, permission=[], root='root') for m in self.models.keys()]

    def _prepare_model(self, model_config: ModelConfig) -> None:
        """Making sure the given model is available locally.

        Args:
            model_config (ModelConfig): Model configuaration of the model to check.
        """
        model_path = model_config.model_id_or_path
        adapter_path = model_config.adapter_path
        quant = model_config.quant
        model_name = model_config.model_name
        if model_name is None:
            model_name = get_model_name(model_path)
            if quant != 'fp16':
                model_name += f'-{quant}'
            if adapter_path:
                model_name += f'-{get_model_name(adapter_path)}'
        if model_name in self.models.keys():
            if self.logger:
                self.logger.error(f'Model "{model_name}" already exists. If there are different configurations with the same model name, please use a different model name.')
                return
            else:
                raise ValueError(f'Model "{model_name}" already exists. If there are different configurations with the same model name, please use a different model name.')
        mlx_path = make_model_exist(model_id_or_path=model_path, quant=quant, revision=model_config.revision, adapter_path=adapter_path)
        mx.metal.clear_cache()
        self.models[model_name] = model_config

    def _switch_model(self, model_name: str) -> None:
        """Switch the model in ram to the given model.

        Args:
            model_name (str): Model name of the model.
        """
        if model_name not in self.models.keys():
            raise ValueError(f'No model named "{model_name}".')
        if model_name != self.current_model:
            del self.model, self.tokenizer, self.cache_manager, self.outlines_tokenizer
            mx.metal.clear_cache()
            model_args = self.models[model_name]._asdict()
            model_args.pop('model_name')
            self.model, self.tokenizer = get_model_and_tokenizer(**model_args)
            self.current_model = model_name
            model_key = get_model_name(self.models[model_name].model_id_or_path)
            self.cache_manager = CacheManager(
                cache_dir=os.path.join(self.cache_dir, model_key),
                token_threshold=self.token_threshold,
                save_similarity_threshold=self.save_sim_threshold,
                max_keep=self.max_keep,
                logger=self.logger
                )
            self.chat_template = ChatTemplate(tokenizer=self.tokenizer._tokenizer)
            if OUTLINE_INSTALLED:
                self.outlines_tokenizer = TransformerTokenizer(tokenizer=self.tokenizer._tokenizer)

    def _get_outlines_processor(self,
            guided_json: Optional[Union[str, dict]] = None,
            guided_choice: Optional[List[str]] = None,
            guided_regex: Optional[str] = None,
            guided_grammar: Optional[str] = None,
            guided_whitespace_pattern: Optional[str] = None
            ) -> Optional[OutlinesLogitsProcessor]:
        logits_processor = None
        if OUTLINE_INSTALLED and (guided_choice) is not None:
            logits_processor=get_choice_processor(guided_choice, self.outlines_tokenizer)

        elif OUTLINE_INSTALLED and (guided_json) is not None:
            logits_processor=get_json_processor(guided_json, self.outlines_tokenizer, guided_whitespace_pattern)

        elif OUTLINE_INSTALLED and (guided_regex) is not None:
            logits_processor=get_regex_processor(guided_regex, self.outlines_tokenizer)
            
        elif OUTLINE_INSTALLED and (guided_grammar) is not None:
            logits_processor = get_grammar_processor(guided_grammar, self.outlines_tokenizer)

        return logits_processor

    def generate(self,
            model: str,
            prompt: Union[str, List[str]],
            completion_type: Literal['text_completion', 'chat.completion'] = 'text_completion', 
            stream: bool = False,
            stop: Optional[List[str]] = None,
            max_tokens: int = 4096,
            n: int = 1,
            temperature: float = 0,
            repetition_penalty: Optional[float] = None,
            repetition_context_size: Optional[int] = 100,
            top_p: float = 1.0,
            min_p: float = 0.0,
            min_tokens_to_keep: int = 1,
            logit_bias: Optional[Dict[int, float]] = None,
            seed: Optional[int] = None,
            guided_json: Optional[Union[str, dict]] = None,
            guided_choice: Optional[List[str]] = None,
            guided_regex: Optional[str] = None,
            guided_grammar: Optional[str] = None,
            guided_whitespace_pattern: Optional[str] = None,
            # chat specific arguments
            tools: Optional[List[Dict[str, Any]]] = None,
            tool_choice: Union[Literal['none','auto', 'required'], Dict[str, Union[str, Dict[str, str]]]] = 'auto',
            **kwargs
        ) -> Union[List[GenerationOutput], Iterator[GenerationOutput]]:
        """Generate text with a model.

        Args:
            model_name (str): Name of the model to use.
            prompt (str): Text prompt for the generation.
            stream (bool, optional): Whether to return a generator for streaming text tokens during generation. Defaults to False.
            stop (Optional[List[str]], optional): List of texts to stop generation. Defaults to None.
            max_tokens (int, optional): Maximum number of tokens to generate. Defaults to 256.
            temperature (float, optional): Temperature that decide the "randomness" of the generation. Defaults to 0.
            repetition_penalty (Optional[float], optional): Penalty for repeated textt during generation. 1.0 being no penalty while higher values gives more penalty. Defaults to None.
            repetition_context_size (Optional[int], optional): Size of previous tokens to consider for repetition penalty. Defaults to 100.
            top_p (float, optional): Top P sampling. Defaults to 1.0.
            min_p (float, optional): Min P sampling. Defaults to 0.0.
            min_tokens_to_keep (int, optional): Minimum number of tokens to keep for sampling. Defaults to 1.
            logit_bias (Optional[Dict[int, float]], optional): Logits bias. Defaults to None.
            prefill_step_size (Optional[int], optional): Batch size for prompt preprocessing. Defaults to None.

        Returns:
            Union[str, Iterator[str]]: Generated text if `stream=False`.

        Yields:
            Iterator[Union[str, Iterator[str]]]: Generator of generated text if `stream=True`.
        """
        self._switch_model(model_name=model)
        prompts = [prompt] if isinstance(prompt, str) else prompt
        prompt_tokens = [remove_bos_duplicates(self.tokenizer.encode(p), bos_token_id=self.tokenizer.bos_token_id) for p in prompts]
        if n != 1:
            prompt_tokens = sum([[p] * n  for p in prompt_tokens], [])
            prompts = sum([[p] * n  for p in prompts], [])
        num_prompts = len(prompts)
        prompt_index = list(range(num_prompts))
        cache_history, cache_ids = self.cache_manager.find_cache(token_ids=prompt_tokens)
        cache_history = CacheHistory(cache=make_empty_cache(model=self.model), token_ids=[]) if cache_history is None else cache_history

        stop = [stop] if isinstance(stop, str) else stop
        
        logits_processor = self._get_outlines_processor(guided_json=guided_json, 
            guided_choice=guided_choice, guided_regex=guided_regex, guided_grammar=guided_grammar, guided_whitespace_pattern=guided_whitespace_pattern)
        
        # Handling tool calls
        tool_stage = None
        tool_name = None
        if tools:
            if tool_choice == 'auto':
                if isinstance(stop, list):
                    stop.append(self.chat_template.tool_start)
                else:
                    stop = [self.chat_template.tool_start]
            elif tool_choice == 'required':
                tool_stage = 'pick_tool'
                tool_names = [get_tool_name(tool) for tool in tools if get_tool_name(tool) is not None]
                logits_processor = self._get_outlines_processor(guided_choice=tool_names)
            elif tool_choice == 'none':
                pass
            else:
                tool_stage = 'gen_args'
                tool_name = tool_choice['function']['name']
                tool_dict = list(filter(lambda x: get_tool_name(x) == tool_name, tools))[0]
                logits_processor = self._get_outlines_processor(guided_json=convert_tool_to_json_schema(tool_dict))

        mx.metal.clear_cache()
        
        gen_kwargs = dict(
            model=self.model,
            tokenizer=self.tokenizer,
            prompt=prompts,
            max_tokens=max_tokens,
            stop=stop,
            temperature=temperature,
            top_p=top_p,
            min_p=min_p,
            min_tokens_to_keep=min_tokens_to_keep,
            repetition_penalty=repetition_penalty,
            repetition_context_size=repetition_context_size,
            cache_history=cache_history,
            prefill_batch_size=self.prefill_step_size,
            logit_bias=logit_bias,
            seed=seed,
            logits_processor=logits_processor,
            verbose=True
        )
        cpl_id = uuid.uuid4().hex
        created = int(dt.now().timestamp())
        if stream:
            # Streaming without tool call
            def stream_output() -> Generator[GenerationOutput, None, None]:
                nonlocal cache_history
                try:
                    stopped = []
                    is_completed = False
                    for go in stream_generate(**gen_kwargs):
                        return_dict = get_return_dict(generation_outputs=go, 
                                id=cpl_id, 
                                model=model, 
                                completion_type=completion_type, 
                                created=created, 
                                stream=stream,
                                indices=prompt_index,
                                tool_start=self.chat_template.tool_start
                            )
                        for g in return_dict:
                            index = g['choices'][0]['index']
                            finish_reason = g['choices'][0]['finish_reason']
                            if finish_reason in ['stop', 'tool_call_start', 'tool_calls']:
                                stopped.append(index)
                            if (index not in stopped) or (finish_reason in ['stop', 'tool_call_start', 'tool_calls']):
                                if finish_reason not in ['stop', 'length']:
                                    g['choices'][0]['finish_reason'] = None
                                yield g
                    is_completed = True
                finally:
                    # To avoid weird behavior if the competion is not completed
                    if is_completed:
                        token_ids = [g.token_ids.tolist() for g in go]
                        new_ch = CacheHistory(cache=cache_history.cache, token_ids=token_ids)
                        self.cache_manager.save_cache(cache=new_ch)
                        del cache_history, new_ch
                    mx.metal.clear_cache()
            
            # Streaming with tool call
            def tool_stream_output():
                if num_prompts != 1:
                    raise Exception(f'Tool calling for more than one chain of messages is not supported.')
                if completion_type != 'chat.completion':
                    raise Exception('Tool calling only supported for chat completion.')
                
                nonlocal cache_history, tool_stage, tool_name
                try:
                    is_completed = False
                    text = ''
                    call_id = uuid.uuid4().hex[:8]

                    # Allow generation if tool_choice is 'auto'
                    if tool_choice == 'auto':
                        for go in stream_generate(**gen_kwargs):
                            g = get_return_dict(generation_outputs=go, 
                                    id=cpl_id, 
                                    model=model, 
                                    completion_type=completion_type, 
                                    created=created, 
                                    stream=stream,
                                    indices=prompt_index,
                                    tool_start=self.chat_template.tool_start
                                )[0]
                            index = g['choices'][0]['index']
                            finish_reason = g['choices'][0]['finish_reason']
                            if finish_reason is None or (finish_reason in ('stop', 'length')):
                                text += g['choices'][0]['delta']['content']
                                yield g
                            elif finish_reason == 'tool_call_start':
                                tool_stage = 'pick_tool'
                                g['choices'][0]['finish_reason'] = None
                                # update the generation kwargs here
                                gen_kwargs['prompt'][0] += text + self.chat_template.tool_start + '{"function": {' + '"name": "'
                                token_ids=[go[0].token_ids.tolist()]
                                self.cache_manager.save_cache(cache=CacheHistory(cache=cache_history.cache, token_ids=token_ids))
                                mx.metal.clear_cache()
                                cache_history, cache_ids = self.cache_manager.find_cache([
                                    remove_bos_duplicates(self.tokenizer.encode(gen_kwargs['prompt'][0]), bos_token_id=self.tokenizer.bos_token_id)
                                ])
                                cache_history = CacheHistory(cache=make_empty_cache(model=self.model), token_ids=[]) if cache_history is None else cache_history
                                gen_kwargs['cache_history'] = cache_history
                                tool_names = [get_tool_name(tool) for tool in tools if get_tool_name(tool) is not None]
                                gen_kwargs['logits_processor'] = self._get_outlines_processor(guided_choice=tool_names)
                                yield g

                    # stage 2, generate tool name
                    if tool_stage == 'pick_tool':
                        go = generate(**gen_kwargs)
                        finish_reason = go[0].finish_reason
                        tool_name = go[0].text
                        if finish_reason == 'stop':
                            tool_stage = 'gen_args'
                            gen_kwargs['prompt'][0] += tool_name + '", arguments": '
                            token_ids=[go[0].token_ids.tolist()]
                            self.cache_manager.save_cache(cache=CacheHistory(cache=cache_history.cache, token_ids=token_ids))
                            mx.metal.clear_cache()
                            # Find cache again
                            cache_history, cache_ids = self.cache_manager.find_cache([
                                remove_bos_duplicates(self.tokenizer.encode(gen_kwargs['prompt'][0]), bos_token_id=self.tokenizer.bos_token_id)
                            ])
                            cache_history = CacheHistory(cache=make_empty_cache(model=self.model), token_ids=[]) if cache_history is None else cache_history
                            gen_kwargs['cache_history'] = cache_history
                            tool_dict = list(filter(lambda x: get_tool_name(x) == tool_name, tools))[0]
                            gen_kwargs['logits_processor'] = self._get_outlines_processor(guided_json=convert_tool_to_json_schema(tool_dict))

                    # stage 3, generate arugments
                    if tool_stage == 'gen_args':
                        go = generate(**gen_kwargs)
                        finish_reason = go[0].finish_reason
                        tool_arguments = go[0].text
                        return_dict = get_return_dict(
                            generation_outputs=go,
                            id=cpl_id, 
                            model=model, 
                            completion_type=completion_type, 
                            created=created, 
                            stream=stream,
                            indices=prompt_index,
                            tool_start=self.chat_template.tool_start,
                            tool_call_id=call_id,
                            tool_name=tool_name,
                            tool_arguments=tool_arguments
                        )
                        yield return_dict[0]
                    is_completed = True

                finally:
                    # To avoid weird behavior if the competion is not completed
                    if is_completed:
                        token_ids = [g.token_ids.tolist() for g in go]
                        new_ch = CacheHistory(cache=cache_history.cache, token_ids=token_ids)
                        self.cache_manager.save_cache(cache=new_ch)
                        del cache_history, new_ch
                    mx.metal.clear_cache()

            return stream_output() if not tools or tool_choice == 'none' else tool_stream_output()

        else:
            call_id = uuid.uuid4().hex[:8]
            tool_name = None
            tool_arguments = None
            if not tools or tool_choice in ['none', 'auto']:
                go = generate(**gen_kwargs)
                token_ids = [g.token_ids.tolist() for g in go]
                text = go[0].text
                finish_reason = go[0].finish_reason
                if (finish_reason == 'stop') and (go[0].stop_text == self.chat_template.tool_start):
                    tool_stage = 'pick_tool'
                    # update the generation kwargs here
                    gen_kwargs['prompt'][0] += text + self.chat_template.tool_start + '{"function": {' + '"name": "'
                    token_ids=[go[0].token_ids.tolist()]
                    self.cache_manager.save_cache(cache=CacheHistory(cache=cache_history.cache, token_ids=token_ids))
                    mx.metal.clear_cache()
                    cache_history, cache_ids = self.cache_manager.find_cache([
                        remove_bos_duplicates(self.tokenizer.encode(gen_kwargs['prompt'][0]), bos_token_id=self.tokenizer.bos_token_id)
                    ])
                    cache_history = CacheHistory(cache=make_empty_cache(model=self.model), token_ids=[]) if cache_history is None else cache_history
                    gen_kwargs['cache_history'] = cache_history
                    tool_names = [get_tool_name(tool) for tool in tools if get_tool_name(tool) is not None]
                    gen_kwargs['logits_processor'] = self._get_outlines_processor(guided_choice=tool_names)
                else:
                    call_id = None

            # stage 2, generate tool name
            if tool_stage == 'pick_tool':
                go = generate(**gen_kwargs)
                finish_reason = go[0].finish_reason
                tool_name = go[0].text
                if finish_reason == 'stop':
                    tool_stage = 'gen_args'
                    gen_kwargs['prompt'][0] += tool_name + '", arguments": '
                    token_ids=[go[0].token_ids.tolist()]
                    self.cache_manager.save_cache(cache=CacheHistory(cache=cache_history.cache, token_ids=token_ids))
                    mx.metal.clear_cache()
                    # Find cache again
                    cache_history, cache_ids = self.cache_manager.find_cache([
                        remove_bos_duplicates(self.tokenizer.encode(gen_kwargs['prompt'][0]), bos_token_id=self.tokenizer.bos_token_id)
                    ])
                    cache_history = CacheHistory(cache=make_empty_cache(model=self.model), token_ids=[]) if cache_history is None else cache_history
                    gen_kwargs['cache_history'] = cache_history
                    tool_dict = list(filter(lambda x: get_tool_name(x) == tool_name, tools))[0]
                    gen_kwargs['logits_processor'] = self._get_outlines_processor(guided_json=convert_tool_to_json_schema(tool_dict))

            # stage 3, generate arugments
            if tool_stage == 'gen_args':
                go = generate(**gen_kwargs)
                finish_reason = go[0].finish_reason
                tool_arguments = go[0].text

            token_ids=[go[0].token_ids.tolist()]
            self.cache_manager.save_cache(cache=CacheHistory(cache=cache_history.cache, token_ids=token_ids))
            del cache_history
            mx.metal.clear_cache()

            return get_return_dict(go, id=cpl_id, model=model, 
                completion_type=completion_type, created=created, stream=False, 
                tool_start=self.chat_template.tool_start, tool_call_id=call_id, tool_name=tool_name, tool_arguments=tool_arguments)
        
    def chat_generate(self,
            model: str,
            messages: Union[List[Dict[str, Any]], List[List[Dict[str, Any]]]],
            stream: bool = False,
            stop: Optional[List[str]] = None,
            max_tokens: int = 4096,
            n: int = 1,
            temperature: float = 0,
            repetition_penalty: Optional[float] = None,
            repetition_context_size: Optional[int] = 100,
            top_p: float = 1.0,
            min_p: float = 0.0,
            min_tokens_to_keep: int = 1,
            logit_bias: Optional[Dict[int, float]] = None,
            seed: Optional[int] = None,
            guided_json: Optional[Union[str, dict]] = None,
            guided_choice: Optional[List[str]] = None,
            guided_regex: Optional[str] = None,
            guided_grammar: Optional[str] = None,
            guided_whitespace_pattern: Optional[str] = None,
            tools: Optional[List[Dict[str, Any]]] = None,
            tool_choice: Union[Literal['none', 'auto', 'required'], Dict[str, Union[str, Dict[str, str]]]] = 'auto',
            **kwargs
        ) -> Union[List[GenerationOutput], Iterator[GenerationOutput]]:
        """Generate text with a model.

        Args:
            model_name (str): Name of the model to use.
            prompt (str): Text prompt for the generation.
            stream (bool, optional): Whether to return a generator for streaming text tokens during generation. Defaults to False.
            stop (Optional[List[str]], optional): List of texts to stop generation. Defaults to None.
            max_tokens (int, optional): Maximum number of tokens to generate. Defaults to 256.
            temperature (float, optional): Temperature that decide the "randomness" of the generation. Defaults to 0.
            repetition_penalty (Optional[float], optional): Penalty for repeated textt during generation. 1.0 being no penalty while higher values gives more penalty. Defaults to None.
            repetition_context_size (Optional[int], optional): Size of previous tokens to consider for repetition penalty. Defaults to 100.
            top_p (float, optional): Top P sampling. Defaults to 1.0.
            min_p (float, optional): Min P sampling. Defaults to 0.0.
            min_tokens_to_keep (int, optional): Minimum number of tokens to keep for sampling. Defaults to 1.
            logit_bias (Optional[Dict[int, float]], optional): Logits bias. Defaults to None.
            prefill_step_size (Optional[int], optional): Batch size for prompt preprocessing. Defaults to None.

        Returns:
            Union[str, Iterator[str]]: Generated text if `stream=False`.

        Yields:
            Iterator[Union[str, Iterator[str]]]: Generator of generated text if `stream=True`.
        """
        self._switch_model(model_name=model)
        if not isinstance(messages[0], list):
            messages = [messages]
        prompt = [self.chat_template.apply_chat_template(messages=msgs, tools=tools, tool_choice=tool_choice) for msgs in messages]
        output = self.generate(
            model=model,
            prompt=prompt,
            completion_type='chat.completion',
            stream=stream,
            stop=stop,
            max_tokens=max_tokens,
            n=n,
            temperature=temperature,
            repetition_penalty=repetition_penalty,
            repetition_context_size=repetition_context_size,
            top_p=top_p,
            min_p=min_p,
            min_tokens_to_keep=min_tokens_to_keep,
            logit_bias=logit_bias,
            seed=seed,
            guided_json=guided_json,
            guided_choice=guided_choice,
            guided_regex=guided_regex,
            guided_grammar=guided_grammar,
            guided_whitespace_pattern=guided_whitespace_pattern,
            tools=tools,
            tool_choice=tool_choice
        )
        return output

    



from .tokenizer_utils import TokenizerWrapper
import mlx.core as mx
import mlx.nn as nn
import time
from .cache_utils import CacheHistory, make_empty_cache
from .sampling_utils import categorical_sampling, min_p_sampling, top_p_sampling
from typing import Optional, Literal, List, NamedTuple, Tuple, Generator, Union, Dict, Any
import warnings
try:
    from outlines.processors.base_logits_processor import OutlinesLogitsProcessor
    from outlines.processors.structured import RegexLogitsProcessor, JSONLogitsProcessor, CFGLogitsProcessor
    from outlines.models.transformers import TransformerTokenizer
    OUTLINE_INSTALLED = True
except:
    warnings.warn('Latest version of outline is not installed. Guided decoding is not enabled.')
    OUTLINE_INSTALLED = False


FINISH_REASON = Optional[Literal['stop', 'length']]

class StopCondition(NamedTuple):
    stop_met: bool
    trim_length: int
    stop_text: Optional[str] = None

class GenerationOutput(NamedTuple):
    text: Optional[str]
    token: Optional[int]
    token_ids: mx.array
    logprobs: mx.array
    finish_reason: FINISH_REASON
    stop_text: Optional[str] = None

def stopping_criteria(
        text: str,
        stop_tuple: List[Tuple[str, int]]
    ) -> StopCondition:
    """Get the stopping condition with stop words and eos token.

    Args:
        text (str): Text to be determined to stop or not.
        stop_tuple (List[Tuple[str, int]]): List of tuple with the stop word string and the length of the string. Must be ordered descendingly by length.

    Returns:
        StopCondition: A status class stating whether the generation should stop and the length of text to trim.
    """
    return next(
        (StopCondition(stop_met=True, trim_length=len(text) - len(text.split(stop)[0]), stop_text=stop)
        for stop, length in stop_tuple if stop in text), StopCondition(stop_met=False, trim_length=0, stop_text=None)
    )

def sequence_overlap(s1: str, s2: str) -> bool:
    """Determine if there is overlapping string between the suffix of the first string and the prefix of the second string.

    Args:
        s1 (str): First string.
        s2 (str): Second string.

    Returns:
        bool: Whether there is overlap.
    """
    max_overlap = min(len(s1), len(s2))
    return any(s1[-i:] == s2[:i] for i in range(1, max_overlap + 1))

def remove_bos_duplicates(token_ids: List[int], bos_token_id: Optional[int]) -> List[int]:
    """Remove duplicates of bos tokens if there are multiple of them in the token id list. This usually happen when the prompt was already formatted by the tokenizer.

    Args:
        token_ids (List[int]): List of token ids.
        bos_token_id (Optional[int]): BOS token id.

    Returns:
        List[int]: List of token ids without bos token duplicates.
    """
    if ((len(token_ids) > 1) and (token_ids[0] == bos_token_id)):
        while token_ids[1] == bos_token_id:
            token_ids = token_ids[:1] + token_ids[2:]
    return token_ids

def get_batches(size: int, batch_size: int) -> List[Tuple[int, int]]:
    """Generate a list of batch ranges for a given size and batch size.

    Args:
        size (int): The total size to be divided into batches.
        batch_size (int): The size of each batch.

    Returns:
        List[Tuple[int, int]]: A list of tuples, where each tuple represents the start and end indices of a batch.
    """
    num_batches = size // batch_size if (size // batch_size) == (size / batch_size) else (size // batch_size) + 1
    batches = [(b * batch_size, min(size, (b + 1) * batch_size)) for b in range(num_batches)]
    return batches

def apply_repetition_penalty(logits: mx.array, tokens: mx.array, penalty: float) -> mx.array:
    """
    Apply repetition penalty to specific logits based on the given context.

    Paper: https://arxiv.org/abs/1909.05858

    Args:
        logits (mx.array): The logits produced by the language model.
        tokens (mx.array): A list of N previous tokens.
        penalty (float): The repetition penalty factor to be applied.

    Returns:
        logits (mx.array): Logits with repetition penalty applied to generated tokens.
    """
    if len(tokens) > 0:
        selected_logits = logits[:, tokens]
        selected_logits = mx.where(
            selected_logits < 0, selected_logits * penalty, selected_logits / penalty
        )
        logits[:, tokens] = selected_logits
    return logits

def generate_step(
        model: nn.Module,
        prompt_ids: mx.array,
        temperature: float = 0.0,
        top_p: float = 1.0,
        min_p: float = 0.0,
        min_tokens_to_keep: int = 1,
        repetition_penalty: Optional[float] = None,
        repetition_context_size: Optional[int] = 20,
        cache_history: Optional[CacheHistory] = None,
        prefill_batch_size: int = 512,
        logit_bias: Optional[Dict[int, float]] = None,
        logits_processor: Optional[OutlinesLogitsProcessor] = None,
        seed: Optional[int] = None,
        verbose: bool = True,
    ) -> Generator[Tuple[mx.array, mx.array], None, None]:
    """
    Generate text tokens step by step using a language model.

    This function processes the prompt, applies repetition penalty, and samples the next token based on the given temperature, top_p, and min_p parameters.
    It also supports logit bias and logits processor for custom logit adjustments.

    Args:
        model (nn.Module): The language model to use for generating tokens.
        prompt_ids (mx.array): The initial token ids for the prompt.
        temperature (float): The temperature parameter for sampling.
        top_p (float): The top_p parameter for nucleus sampling.
        min_p (float): The min_p parameter for min_p sampling.
        min_tokens_to_keep (int): The minimum number of tokens to keep for min_p sampling.
        repetition_penalty (Optional[float]): The repetition penalty factor to apply.
        repetition_context_size (Optional[int]): The context size for repetition penalty.
        cache_history (Optional[CacheHistory]): The cache history to use for caching.
        prefill_batch_size (int): The batch size for prefilling.
        logit_bias (Optional[Dict[int, float]]): The logit bias to apply.
        logits_processor (Optional[OutlinesLogitsProcessor]): The logits processor to use for guided decoding.
        seed (Optional[int]): The random seed for reproducibility.
        verbose (bool): Whether to print progress information.

    Yields:
        Tuple[mx.array, mx.array]: The next token ids and their log probabilities.
    """

    # Create or validate the cache history
    num_prompts = prompt_ids.shape[0]
    prompt_size = prompt_ids.shape[1]
    cache_size = 0
    if cache_history is None:
        cache_history = CacheHistory(cache=make_empty_cache(model), token_ids=[[]] * num_prompts)
    elif cache_history.cache[0].offset == 0:
        pass
    elif cache_history.cache[0].state[0].shape[0] != num_prompts:
        raise ValueError('Cache history number of slots does not match with number of prompts.')
    elif cache_history.cache[0].state[0].shape[2] >= prompt_size:
        raise ValueError('Cache history size larger than prompt size.')
    else:
        cache_size = cache_history.cache[0].state[0].shape[2]
    
    # Trimming prompt ids with cache
    unprocessed_ids = prompt_ids[:, cache_size:]

    # Preparing resources
    num_gen_tokens: int = 0
    rng = mx.arange(num_prompts).reshape((-1, 1))
    if seed is not None:
        mx.random.seed(seed)

    # Function for processing the next token
    if logit_bias:
        indices = mx.array(list(logit_bias.keys()))
        values = mx.array(list(logit_bias.values()))

        def logit_bias_processor(logits):
            logits[:, indices] += values
            return logits
        

    def sample(logits: mx.array) -> Tuple[mx.array, mx.array]:
        logprobs = logits - mx.logsumexp(logits)
        if temperature == 0:
            y = mx.argmax(logits, axis=1).reshape(-1, 1)
        elif top_p > 0 and top_p < 1.0:
            y = top_p_sampling(logits=logits, top_p=top_p, temperature=temperature, rng=rng)
        elif min_p != 0.0:
            y = min_p_sampling(logits=logits, min_p=min_p, rng=rng, min_tokens_to_keep=min_tokens_to_keep, temperature=temperature)
        else:
            y = categorical_sampling(logits=logits, temperature=temperature)
        return y, logprobs

    def _step(y: mx.array) -> Tuple[List[int], mx.array]:
        logits = model(y, cache=cache_history.cache)
        logits = logits[:, -1, :]

        if repetition_penalty:
            tks = prompt_ids[:, -repetition_context_size:] if repetition_context_size is not None else prompt_ids
            logits = apply_repetition_penalty(logits=logits, tokens=tks, penalty=repetition_penalty)

        if logit_bias:
            logits = logit_bias_processor(logits)

        if logits_processor:
            input_ids = prompt_ids[:, -num_gen_tokens:] if num_gen_tokens != 0 else [[]] * num_prompts
            logits = logits_processor(input_ids=input_ids, logits=logits)

        y, logprobs = sample(logits)
        return y, logprobs

    # Prompt processing
    batches = get_batches(unprocessed_ids.shape[1], batch_size=prefill_batch_size)
    if verbose:
        from tqdm import tqdm
        batches = tqdm(batches)
    pp_start = time.perf_counter()
    for b1, b2 in batches:
        if verbose:
            batches.set_description(f'Processing prompt ({b2}/{unprocessed_ids.shape[1]})')

        if b2 != unprocessed_ids.shape[1]: # proccesing without sampling and yielding next token
            model(unprocessed_ids[:, b1:b2], cache=cache_history.cache)
            mx.eval([c.state for c in cache_history.cache])
            mx.metal.clear_cache()

        else: # proccesing with sampling and ygetting next tokens with the last batch
            y, logprobs = _step(unprocessed_ids[:, b1:b2])
            prompt_ids = mx.concat([prompt_ids, y], axis=1)
            num_gen_tokens += 1
    pp_end = time.perf_counter() - pp_start
    if verbose:
        num_tokens = unprocessed_ids.shape[1] * num_prompts
        print(f'Prompt preprocessing time for {num_tokens} tokens: {pp_end:.4}s ({num_tokens/pp_end:.4f} tok/sec)')
    mx.async_eval(y)

    # Generating tokens
    while True:
        next_y, next_logprobs = _step(y)
        mx.async_eval(next_y)
        prompt_ids = mx.concat([prompt_ids, next_y], axis=1)
        num_gen_tokens += 1
        yield y, logprobs
        y, logprobs = next_y, next_logprobs
    
def stream_generate(
        model: nn.Module,
        tokenizer: TokenizerWrapper,
        prompt: Union[str, List[str]],
        max_tokens: int = 512,
        stop: Optional[List[str]]  = None,
        temperature: float = 0.0,
        top_p: float = 1.0,
        min_p: float = 0.0,
        min_tokens_to_keep: int = 1,
        repetition_penalty: Optional[float] = None,
        repetition_context_size: Optional[int] = 20,
        cache_history: Optional[CacheHistory] = None,
        prefill_batch_size: int = 512,
        logit_bias: Optional[Dict[int, float]] = None,
        logits_processor: Optional[OutlinesLogitsProcessor] = None,
        seed: Optional[int] = None,
        verbose: bool = True,
    ) -> Generator[List[GenerationOutput], None, None]:
    """
    Stream generate text tokens using a language model.

    This function tokenizes the prompt, processes the tokens, and generates text step by step.
    It supports various sampling methods, repetition penalty, and logit adjustments.

    Args:
        model (nn.Module): The language model to use for generating tokens.
        tokenizer (TokenizerWrapper): The tokenizer to use for tokenizing the prompt.
        prompt (Union[str, List[str]]): The initial prompt to generate text from.
        max_tokens (int): The maximum number of tokens to generate.
        stop (Optional[List[str]]): A list of stop words or tokens to stop generation.
        temperature (float): The temperature parameter for sampling.
        top_p (float): The top_p parameter for nucleus sampling.
        min_p (float): The min_p parameter for min_p sampling.
        min_tokens_to_keep (int): The minimum number of tokens to keep for min_p sampling.
        repetition_penalty (Optional[float]): The repetition penalty factor to apply.
        repetition_context_size (Optional[int]): The context size for repetition penalty.
        cache_history (Optional[CacheHistory]): The cache history to use for caching.
        prefill_batch_size (int): The batch size for prefilling.
        logit_bias (Optional[Dict[int, float]]): The logit bias to apply.
        logits_processor (Optional[OutlinesLogitsProcessor]): The logits processor to use for guided decoding.
        seed (Optional[int]): The random seed for reproducibility.
        verbose (bool): Whether to print progress information.

    Yields:
        List[GenerationOutput]: A list of generation outputs, each containing the generated text, token, token ids, log probabilities, finish reason, and stop text.
    """
    # Preparing prompts
    if isinstance(prompt, str):
        prompt = [prompt]
    bos_token_id = tokenizer.bos_token_id
    pad_token_id = bos_token_id if tokenizer.pad_token_id is None else tokenizer.pad_token_id
    prompt_ids = [remove_bos_duplicates(token_ids=tokenizer.encode(p), bos_token_id=bos_token_id) for p in prompt]
    prompt_lens = [len(p) for p in prompt_ids]
    max_len = max(prompt_lens)
    prompt_ids = [[pad_token_id] * (max_len - p[1]) + p [0] for p in zip(prompt_ids, prompt_lens)]
    prompt_ids = mx.array(prompt_ids)

    # Preparing stop sequences
    eos_token = tokenizer.eos_token
    stop = [eos_token] if (stop is None) or (logits_processor is not None) else [eos_token] + stop # Ignore stop words if guided decoding is used.
    stop = list(set(stop))
    stop_len = [len(s) for s in stop]
    stop_tuple = list(zip(stop, stop_len))
    stop_tuple.sort(key=lambda x: x[1], reverse=True)

    # Generation
    def process_stop_condition(text: str, 
                               stop: List[str], 
                               stop_condition: StopCondition,
                               is_stop: bool,
                               finish_reason: FINISH_REASON,
                               is_last: bool) -> Tuple[str, str, FINISH_REASON, Optional[str]]:
        if is_stop:
            return '', '', None, None
        elif stop_condition.stop_met:
            return text[:-stop_condition.trim_length], '', 'stop', stop_condition.stop_text
        elif any(sequence_overlap(text, s) for s in stop) and (not is_last):
            return '', text, finish_reason, stop_condition.stop_text
        else:
            return text, '', finish_reason, stop_condition.stop_text


    texts = [''] * prompt_ids.shape[0]
    is_stopped = [False] * prompt_ids.shape[0]
    detokenizer = tokenizer._detokenizer
    detokenizer.reset()
    try:
        if verbose:
            start = None
        for (tokens, logprobs), n in zip(
            generate_step(
                model=model,
                prompt_ids=prompt_ids,
                temperature=temperature,
                top_p=top_p,
                min_p=min_p,
                min_tokens_to_keep=min_tokens_to_keep,
                repetition_penalty=repetition_penalty,
                repetition_context_size=repetition_context_size,
                cache_history=cache_history,
                prefill_batch_size=prefill_batch_size,
                logit_bias=logit_bias,
                logits_processor=logits_processor,
                seed=seed,
                verbose=verbose
            ),
            range(max_tokens)
        ):
            if verbose:
                start = time.perf_counter() if n == 0 else start
            finish_reason = None if (n + 1) != max_tokens else 'length'
            new_tokens = tokens.tolist()
            detokenizer.add_tokens(new_tokens)
            # finalize for last token
            if (n + 1) == max_tokens:
                detokenizer.finalize()
            new_token_str = detokenizer.last_segments
            prompt_ids = mx.concat([prompt_ids, tokens], axis=1)
            texts = [t[0] + t[1] for t in zip(texts, new_token_str)]
            stop_conditions = [stopping_criteria(t, stop_tuple=stop_tuple) for t in texts]
            outputs = [process_stop_condition(t, stop, sc, s, finish_reason, finish_reason=='length') for t, sc, s in zip(texts, stop_conditions, is_stopped)]
            out_texts, texts, finish_reasons, stop_texts = [list(ls) for ls in zip(*outputs)]
            is_stopped = [s or (fr == 'stop') for s, fr in zip(is_stopped, finish_reasons)]
            gen_outputs = zip(out_texts, new_tokens, prompt_ids, logprobs, finish_reasons, prompt_lens, stop_texts)
            gen_outputs = [GenerationOutput(text=go[0], 
                        token=go[1][0], 
                        token_ids=go[2][-(go[5] + n +1):], 
                        logprobs=go[3][None], 
                        finish_reason=go[4],
                        stop_text=go[6]) for go in gen_outputs]
            yield gen_outputs
            if all(is_stopped):
                break
        
    finally:
        if verbose:
            end = time.perf_counter() - start
            num_tokens = (n + 1) * len(prompt_ids)
            print(f'Number of tokens generated: {num_tokens}; Generation time: {end}s ({(num_tokens / end):.4f} tok/sec)')
        mx.metal.clear_cache()
    
def generate(
        model: nn.Module,
        tokenizer: TokenizerWrapper,
        prompt: Union[str, List[str]],
        max_tokens: int = 512,
        stop: Optional[List[str]]  = None,
        temperature: float = 0.0,
        top_p: float = 1.0,
        min_p: float = 0.0,
        min_tokens_to_keep: int = 1,
        repetition_penalty: Optional[float] = None,
        repetition_context_size: Optional[int] = 20,
        cache_history: Optional[CacheHistory] = None,
        prefill_batch_size: int = 512,
        logit_bias: Optional[Dict[int, float]] = None,
        logits_processor: Optional[OutlinesLogitsProcessor] = None,
        seed: Optional[int] = None,
        verbose: bool = True,
    ) -> List[GenerationOutput]:
    """
    Generate text tokens using a language model.

    This function tokenizes the prompt, processes the tokens, and generates text step by step.
    It supports various sampling methods, repetition penalty, and logit adjustments.

    Args:
        model (nn.Module): The language model to use for generating tokens.
        tokenizer (TokenizerWrapper): The tokenizer to use for tokenizing the prompt.
        prompt (Union[str, List[str]]): The initial prompt to generate text from.
        max_tokens (int): The maximum number of tokens to generate.
        stop (Optional[List[str]]): A list of stop words or tokens to stop generation.
        temperature (float): The temperature parameter for sampling.
        top_p (float): The top_p parameter for nucleus sampling.
        min_p (float): The min_p parameter for min_p sampling.
        min_tokens_to_keep (int): The minimum number of tokens to keep for min_p sampling.
        repetition_penalty (Optional[float]): The repetition penalty factor to apply.
        repetition_context_size (Optional[int]): The context size for repetition penalty.
        cache_history (Optional[CacheHistory]): The cache history to use for caching.
        prefill_batch_size (int): The batch size for prefilling.
        logit_bias (Optional[Dict[int, float]]): The logit bias to apply.
        logits_processor (Optional[OutlinesLogitsProcessor]): The logits processor to use for guided decoding.
        seed (Optional[int]): The random seed for reproducibility.
        verbose (bool): Whether to print progress information.

    Returns:
        List[GenerationOutput]: A list of generation outputs, each containing the generated text, token, token ids, log probabilities, finish reason, and stop text.
    """
    # Preparing prompts
    if isinstance(prompt, str):
        prompt = [prompt]
    bos_token_id = tokenizer.bos_token_id
    pad_token_id = bos_token_id if tokenizer.pad_token_id is None else tokenizer.pad_token_id
    prompt_ids = [remove_bos_duplicates(token_ids=tokenizer.encode(p), bos_token_id=bos_token_id) for p in prompt]
    prompt_lens = [len(p) for p in prompt_ids]
    max_len = max(prompt_lens)
    prompt_ids = [[pad_token_id] * (max_len - p[1]) + p [0] for p in zip(prompt_ids, prompt_lens)]
    prompt_ids = mx.array(prompt_ids)

    # Preparing stop sequences
    eos_token = tokenizer.eos_token
    stop = [eos_token] if (stop is None) or (logits_processor is not None) else [eos_token] + stop # Ignore stop words if guided decoding is used.
    stop = list(set(stop))
    stop_len = [len(s) for s in stop]
    stop_tuple = list(zip(stop, stop_len))
    stop_tuple.sort(key=lambda x: x[1], reverse=True)

    # Generation
    texts = [''] * prompt_ids.shape[0]
    is_stopped = [False] * prompt_ids.shape[0]
    stop_texts = [None] * prompt_ids.shape[0]
    detokenizer = tokenizer._detokenizer
    detokenizer.reset()
    try:
        if verbose:
            start = None
        for (tokens, logprobs), n in zip(
            generate_step(
                model=model,
                prompt_ids=prompt_ids,
                temperature=temperature,
                top_p=top_p,
                min_p=min_p,
                min_tokens_to_keep=min_tokens_to_keep,
                repetition_penalty=repetition_penalty,
                repetition_context_size=repetition_context_size,
                cache_history=cache_history,
                prefill_batch_size=prefill_batch_size,
                logit_bias=logit_bias,
                logits_processor=logits_processor,
                seed=seed,
                verbose=verbose
            ),
            range(max_tokens)
        ):
            if verbose:
                start = time.perf_counter() if n == 0 else start
            new_tokens = tokens.tolist()
            detokenizer.add_tokens(new_tokens)
            # finalize for last token
            if (n + 1) == max_tokens:
                detokenizer.finalize()
            new_token_str = detokenizer.last_segments
            prompt_ids = mx.concat([prompt_ids, tokens], axis=1)
            texts = [t[0] + t[1] if not t[2] else t[0] for t in zip(texts, new_token_str, is_stopped)]
            stop_conditions = [stopping_criteria(t, stop_tuple=stop_tuple) for t in texts]
            stop_texts = [sc.stop_text if sc.stop_met and (not s) else st for s, sc, st in zip(is_stopped, stop_conditions, stop_texts)]
            is_stopped = [s or sc.stop_met for s, sc in zip(is_stopped, stop_conditions)]
            if all(is_stopped):
                break
        if verbose:
            end = time.perf_counter() - start
            num_tokens = (n + 1) * len(prompt_ids)
            print(f'Number of tokens generated: {num_tokens}; Generation time: {end}s ({(num_tokens / end):.4f} tok/sec)')

        texts = [t[:-sc.trim_length] if s else t for t, sc, s in zip(texts, stop_conditions, is_stopped)]
        finish_reasons = ['stop' if s else 'length' for s in is_stopped]
        gen_outputs = zip(texts, new_tokens, prompt_ids, logprobs, finish_reasons, prompt_lens, stop_texts)
        gen_outputs = [GenerationOutput(text=go[0], token=go[1][0], token_ids=go[2][-(go[5] + n + 1):], logprobs=go[3][None], finish_reason=go[4], stop_text=go[6]) for go in gen_outputs]
        return gen_outputs
        
    finally:
        mx.metal.clear_cache()


# Guided decoding logits processors
def get_regex_processor(regex_str: str, tokenizer: TransformerTokenizer) -> RegexLogitsProcessor:
    return RegexLogitsProcessor(regex_str, tokenizer=tokenizer)

def get_json_processor(schema: Union[Dict[str, Any], str], tokenizer: TransformerTokenizer, whitespace_pattern: Optional[str] = None) -> JSONLogitsProcessor:
    return JSONLogitsProcessor(schema=schema, tokenizer=tokenizer, whitespace_pattern=whitespace_pattern)

def get_choice_processor(choices: List[str], tokenizer: TransformerTokenizer) -> RegexLogitsProcessor:
    regex_str = r"(" + r"|".join(choices) + r")"
    return get_regex_processor(regex_str=regex_str, tokenizer=tokenizer)

def get_grammar_processor(cfg_str: str, tokenizer: TransformerTokenizer) -> CFGLogitsProcessor:
    return CFGLogitsProcessor(cfg_str=cfg_str, tokenizer=tokenizer)



# adapted from mlx-lm
import json
from functools import partial
from typing import List, Optional

from transformers import AutoTokenizer, PreTrainedTokenizer
from huggingface_hub import hf_hub_download

REPLACEMENT_CHAR = "\ufffd"


def _remove_space(x):
    if x and x[0] == " ":
        return x[1:]
    return x


class NaiveDetokenizer:

    def __init__(self, tokenizer: PreTrainedTokenizer) -> None:
        self._tokenizer = tokenizer
        self.reset()

    def reset(self, num_seqs: Optional[int] = None) -> None:
        self.num_seqs = num_seqs
        self.texts = [] if self.num_seqs is None else [''] * self.num_seqs
        self.tokens = [] if self.num_seqs is None else [[]] * self.num_seqs
        self.current_tokens = [] if self.num_seqs is None else [[]] * self.num_seqs
        self.final_segments = None

    def add_tokens(self, token_ids: List[List[int]]) -> None:
        if self.num_seqs is None:
            self.reset(num_seqs=len(token_ids))
        elif len(token_ids) != self.num_seqs:
            raise Exception('Number of tokens does not match with number of existing sequences.')
        self.current_tokens = [tids + nids for tids, nids in zip(self.current_tokens, token_ids)]

    def finalize(self) -> None:
        new_texts = self._tokenizer.batch_decode(self.current_tokens)
        new_texts = [nt.rstrip(REPLACEMENT_CHAR) for nt in new_texts]
        self.tokens = [t + ct for t, ct in zip(self.tokens, self.current_tokens)]
        self.texts = [t + nt for t, nt in zip(self.texts, new_texts)]
        self.current_tokens = [[]] * self.num_seqs
        self.final_segments = new_texts        

    @property
    def last_segments(self) -> List[str]:
        if self.final_segments is not None:
            return self.final_segments
        new_texts = self._tokenizer.batch_decode(self.current_tokens)
        with_repl = [newt.endswith(REPLACEMENT_CHAR) for newt in new_texts]
        bundle = [(ot if wr else ot + nt, nt if wr else [], os if wr else os + ns, '' if wr else ns) for ot, nt, os, ns, wr in zip(self.tokens, self.current_tokens, self.texts, new_texts, with_repl)]
        tokens, current_tokens, texts, new_segments = list(zip(*bundle))
        self.tokens = list(tokens)
        self.current_tokens = list(current_tokens)
        self.texts = list(texts)
        new_segments = list(new_segments)
        return new_segments

class SPMDetokenizer:
    """A streaming detokenizer for SPM models.

    It adds tokens to the text if the next token starts with the special SPM
    underscore which results in linear complexity.
    """

    def __init__(self, tokenizer: PreTrainedTokenizer, trim_space=True):
        self.trim_space = trim_space
        self.eos_token = tokenizer.eos_token

        # Extract the tokens in a list from id to text
        self.tokenmap = [""] * (max(tokenizer.vocab.values()) + 1)
        for value, tokenid in tokenizer.vocab.items():
            self.tokenmap[tokenid] = value

        # Replace bytes with their value
        for i in range(len(self.tokenmap)):
            if self.tokenmap[i].startswith("<0x"):
                self.tokenmap[i] = chr(int(self.tokenmap[i][3:5], 16))

        self.reset()

    def reset(self, num_seqs: Optional[int] = None) -> None:
        self.num_seqs = num_seqs
        self.offsets = [] if self.num_seqs is None else [0] * self.num_seqs
        self._unflushed = [] if self.num_seqs is None else [''] * self.num_seqs
        self.texts = [] if self.num_seqs is None else [''] * self.num_seqs
        self.tokens = [] if self.num_seqs is None else [[]] * self.num_seqs

    def add_tokens(self, token_ids: List[List[int]]) -> None:
        if self.num_seqs is None:
            self.reset(num_seqs=len(token_ids))
        elif len(token_ids) != self.num_seqs:
            raise Exception('Number of tokens does not match with number of existing sequences.')
        vs = [self.tokenmap[token[0]] for token in token_ids]
        for i, v in enumerate(vs):
            if v[0] == "\u2581":
                if self.texts[i] or not self.trim_space:
                    self.texts[i] += self._unflushed[i].replace("\u2581", " ")
                else:
                    self.texts[i] = _remove_space(self._unflushed[i].replace("\u2581", " "))
                self._unflushed[i] = v
            elif self.eos_token == v:
                if self.texts[i] or not self.trim_space:
                    self.texts[i] += self._unflushed[i].replace("\u2581", " ") + v
                else:
                    self.texts[i] = _remove_space(self._unflushed[i].replace("\u2581", " ")) + v
                self._unflushed[i] = ''
            else:
                self._unflushed[i] += v

    def finalize(self):
        for i in range(self.num_seqs):
            if self.texts[i] or not self.trim_space:
                self.texts[i] += self._unflushed[i].replace("\u2581", " ")
            else:
                self.texts[i] = _remove_space(self._unflushed[i].replace("\u2581", " "))
            self._unflushed[i] = ""

    @property
    def last_segments(self):
        """Return the last segment of readable text since last time this property was accessed."""
        texts = self.texts
        segments = []
        for i, text in enumerate(texts):
            if text and text[-1] != REPLACEMENT_CHAR:
                segments.append(text[self.offsets[i] :])
                self.offsets[i] = len(text)
            else:
                segments.append('')
        return segments



class BPEDetokenizer:
    """A streaming detokenizer for OpenAI style BPE models.

    It adds tokens to the text if the next token starts with a space similar to
    the SPM detokenizer.
    """

    _byte_decoder = None
    _space_matches = (".", "?", "!", ",", "'", "n't", "'m", "'s", "'ve", "'re")

    def __init__(self, tokenizer):

        self.clean_spaces = tokenizer.clean_up_tokenization_spaces

        # Extract the tokens in a list from id to text
        self.tokenmap = [None] * len(tokenizer.vocab)
        for value, tokenid in tokenizer.vocab.items():
            self.tokenmap[tokenid] = value

        self.reset()

        # Make the BPE byte decoder from
        # https://github.com/openai/gpt-2/blob/master/src/encoder.py
        self.make_byte_decoder()

    def reset(self, num_seqs: Optional[int] = None) -> None:
        self.num_seqs = num_seqs
        self.offsets = [] if self.num_seqs is None else [0] * self.num_seqs
        self._unflushed = [] if self.num_seqs is None else [''] * self.num_seqs
        self.texts = [] if self.num_seqs is None else [''] * self.num_seqs
        self.tokens = [] if self.num_seqs is None else [[]] * self.num_seqs

    def _maybe_trim_space(self, current_text, i):
        if len(current_text) == 0:
            return current_text
        elif current_text[0] != " ":
            return current_text
        elif not self.texts[i]:
            return current_text[1:]
        elif self.clean_spaces and current_text[1:].startswith(self._space_matches):
            return current_text[1:]
        return current_text

    def add_tokens(self, token_ids: List[List[int]]) -> None:
        if self.num_seqs is None:
            self.reset(num_seqs=len(token_ids))
        elif len(token_ids) != self.num_seqs:
            raise Exception('Number of tokens does not match with number of existing sequences.')
        vs = [self.tokenmap[token[0]] for token in token_ids]
        for i, v in enumerate(vs):
            if self._byte_decoder[v[0]] == 32:
                current_text = bytearray(
                    self._byte_decoder[c] for c in self._unflushed[i]
                ).decode("utf-8")
                self.texts[i] += self._maybe_trim_space(current_text, i)
                self._unflushed[i] = v
            else:
                self._unflushed[i] += v

    def finalize(self):
        for i in range(self.num_seqs):
            current_text = bytearray(self._byte_decoder[c] for c in self._unflushed[i]).decode(
                "utf-8"
            )
            self.texts[i] += self._maybe_trim_space(current_text, i)
            self._unflushed[i] = ""

    @property
    def last_segments(self):
        """Return the last segment of readable text since last time this property was accessed."""
        texts = self.texts
        segments = []
        for i, text in enumerate(texts):
            if text and text[-1] != REPLACEMENT_CHAR:
                segments.append(text[self.offsets[i] :])
                self.offsets[i] = len(text)
            else:
                segments.append('')
        return segments

    @classmethod
    def make_byte_decoder(cls):
        """See https://github.com/openai/gpt-2/blob/master/src/encoder.py for the rationale."""
        if cls._byte_decoder is not None:
            return

        char_to_bytes = {}
        limits = [
            0,
            ord("!"),
            ord("~") + 1,
            ord("¡"),
            ord("¬") + 1,
            ord("®"),
            ord("ÿ") + 1,
        ]
        n = 0
        for i, (start, stop) in enumerate(zip(limits, limits[1:])):
            if i % 2 == 0:
                for b in range(start, stop):
                    char_to_bytes[chr(2**8 + n)] = b
                    n += 1
            else:
                for b in range(start, stop):
                    char_to_bytes[chr(b)] = b
        cls._byte_decoder = char_to_bytes


class TokenizerWrapper:
    """A wrapper that combines an HF tokenizer and a detokenizer.

    Accessing any attribute other than the ``detokenizer`` is forwarded to the
    huggingface tokenizer.
    """

    def __init__(self, tokenizer, detokenizer_class=NaiveDetokenizer):
        self._tokenizer = tokenizer
        self._detokenizer_class = detokenizer_class
        self._detokenizer = detokenizer_class(tokenizer)

    def __getattr__(self, attr):
        if attr == "detokenizer":
            return self._detokenizer
        elif attr.startswith("_"):
            return self.__getattribute__(attr)
        else:
            return getattr(self._tokenizer, attr)

    def __setattr__(self, attr, value):
        if attr == "detokenizer":
            raise AttributeError("Cannot set the detokenizer.")
        elif attr.startswith("_"):
            super().__setattr__(attr, value)
        else:
            setattr(self._tokenizer, attr, value)



def _match(a, b):
    if type(a) != type(b):
        return False
    if isinstance(a, dict):
        return len(a) == len(b) and all(k in b and _match(a[k], b[k]) for k in a)
    if isinstance(a, list):
        return len(a) == len(b) and all(_match(ai, bi) for ai, bi in zip(a, b))

    return a == b


def _is_spm_decoder(decoder):
    _target_description = {
        "type": "Sequence",
        "decoders": [
            {"type": "Replace", "pattern": {"String": "▁"}, "content": " "},
            {"type": "ByteFallback"},
            {"type": "Fuse"},
            {"type": "Strip", "content": " ", "start": 1, "stop": 0},
        ],
    }
    return _match(_target_description, decoder)


def _is_spm_decoder_no_space(decoder):
    _target_description = {
        "type": "Sequence",
        "decoders": [
            {"type": "Replace", "pattern": {"String": "▁"}, "content": " "},
            {"type": "ByteFallback"},
            {"type": "Fuse"},
        ],
    }
    return _match(_target_description, decoder)


def _is_bpe_decoder(decoder):
    return isinstance(decoder, dict) and decoder.get("type", None) == "ByteLevel"


def load_tokenizer(model_path, tokenizer_config_extra={}):
    """Load a huggingface tokenizer and try to infer the type of streaming
    detokenizer to use.

    Note, to use a fast streaming tokenizer, pass a local file path rather than
    a Hugging Face repo ID.
    """
    detokenizer_class = NaiveDetokenizer

    tokenizer_file = model_path / "tokenizer.json"
    if not tokenizer_file.exists():
        tokenizer_file = hf_hub_download(repo_id=str(model_path), filename='tokenizer.json')
    with open(tokenizer_file, "r") as fid:
        tokenizer_content = json.load(fid)
    if "decoder" in tokenizer_content:
        if _is_spm_decoder(tokenizer_content["decoder"]):
            detokenizer_class = SPMDetokenizer
        elif _is_spm_decoder_no_space(tokenizer_content["decoder"]):
            detokenizer_class = partial(SPMDetokenizer, trim_space=False)
        # elif _is_bpe_decoder(tokenizer_content["decoder"]):
        #     detokenizer_class = BPEDetokenizer

    return TokenizerWrapper(
        AutoTokenizer.from_pretrained(model_path, **tokenizer_config_extra),
        detokenizer_class,
    )